import numpy as np
import torch

from unittest import TestCase
from ezyrb import AE

snapshots = np.load('tests/test_datasets/p_snapshots.npy').T


class TestAE(TestCase):
    def test_constructor_empty(self):
        AE([20, 2], [2, 20], torch.nn.ReLU, torch.nn.ReLU, 20)

    def test_wrong_constructor(self):
        with self.assertRaises(ValueError):
            AE([20, 2], [5, 20], torch.nn.ReLU, torch.nn.ReLU, 20)

    def test_reconstruction(self):
        f = torch.nn.Softplus
        ae = AE([400, 20, 2], [2, 20, 400], f(), f(), 1e-5)
        ae.fit(snapshots)
        snapshots_ = ae.inverse_transform(ae.transform(snapshots))
        rerr = np.linalg.norm(snapshots_ - snapshots)/np.linalg.norm(snapshots)
        assert rerr < 5e-3

    def test_decode_encode(self):
        f = torch.nn.Softplus
        low_dim = 5
        ae = AE([400, low_dim], [low_dim, 400], f(), f(), 20)
        ae.fit(snapshots)
        reduced_snapshots = ae.transform(snapshots)
        assert reduced_snapshots.shape[0] == low_dim
        expanded_snapshots = ae.inverse_transform(reduced_snapshots)
        assert expanded_snapshots.shape[0] == snapshots.shape[0]

    def test_optimizer(self):
        f = torch.nn.Softplus
        low_dim = 5
        ae = AE([400, low_dim], [low_dim, 400], f(), f(), 20)
        ae.fit(snapshots)
        assert ae.optimizer == torch.optim.Adam
        ae = AE([200, 100, 10], [10, 100, 200], torch.nn.Tanh(), torch.nn.Tanh(), 10)
        ae.fit(snapshots)
        assert ae.optimizer == torch.optim.Adam

    def test_optimizer_doublefit(self):
        f = torch.nn.Softplus
        low_dim = 5
        ae = AE([400, low_dim], [low_dim, 400], f(), f(), 20)
        ae.fit(snapshots)
        ae.fit(snapshots)



"""
Module for Artificial Neural Network (ANN) Prediction.
"""

import torch
from torch import nn
import numpy as np
from pycompss.api.task import task
from pycompss.api.parameter import INOUT, IN
from .approximation import Approximation

class ANN(Approximation):
    """
    Feed-Forward Artifical Neural Network class (ANN).

    :param list layers: ordered list with the number of neurons of each
        hidden layer.
    :param torch.nn.modules.activation function: activation function at each
        layer, except for the output layer at with Identity is considered by
        default. A single activation function can be passed or a list of
        them of length equal to the number of hidden layers.
    :param list stop_training: list with the maximum number of training
        iterations (int) and/or the desired tolerance on the training loss
        (float).
    :param torch.nn.Module loss: loss definition (Mean Squared if not
        given).

    :Example:
        >>> import ezyrb
        >>> import numpy as np
        >>> import torch.nn as nn
        >>> x = np.random.uniform(-1, 1, size =(4, 2))
        >>> y = np.array([np.sin(x[:, 0]), np.cos(x[:, 1]**3)]).T
        >>> ann = ezyrb.ANN([10, 5], nn.Tanh(), [20000,1e-5])
        >>> ann.fit(x, y)
        >>> y_pred = ann.predict(x)
        >>> print(y)
        >>> print(y_pred)
        >>> print(len(ann.loss_trend))
        >>> print(ann.loss_trend[-1])
    """
    def __init__(self, layers, function, stop_training, loss=None):

        if loss is None:
            loss = torch.nn.MSELoss()

        if not isinstance(function, list):  # Single activation function passed
            function = [function] * (len(layers))

        if not isinstance(stop_training, list):
            stop_training = [stop_training]

        self.layers = layers
        self.function = function
        self.loss = loss
        self.stop_training = stop_training

        self.loss_trend = []
        self.model = None
        self.optimizer = None

    def _convert_numpy_to_torch(self, array):
        """
        Converting data type.

        :param numpy.ndarray array: input array.
        :return: the tensorial counter-part of the input array.
        :rtype: torch.Tensor.
        """
        return torch.from_numpy(array).float()

    def _convert_torch_to_numpy(self, tensor):
        """
        Converting data type.

        :param torch.Tensor tensor: input tensor.
        :return: the vectorial counter-part of the input tensor.
        :rtype: numpy.ndarray.
        """
        return tensor.detach().numpy()

    def _build_model(self, points, values):
        """
        Build the torch model.
        Considering the number of neurons per layer (self.layers), a
        feed-forward NN is defined:
        -  activation function from layer i>=0 to layer i+1:
           self.function[i]; activation function at the output layer:
           Identity (by default).
        :param numpy.ndarray points: the coordinates of the given (training)
        points.
        :param numpy.ndarray values: the (training) values in the points.
        """
        layers = self.layers.copy()
        layers.insert(0, points.shape[1])
        layers.append(values.shape[1])
        layers_torch = []
        for i in range(len(layers) - 2):
            layers_torch.append(nn.Linear(layers[i], layers[i + 1]))
            layers_torch.append(self.function[i])
        layers_torch.append(nn.Linear(layers[-2], layers[-1]))
        self.model = nn.Sequential(*layers_torch)

    @task(target_direction=INOUT)
    def fit(self, points, values):
        """
        Build the ANN given 'points' and 'values' and perform training.

        Training procedure information:
            -  optimizer: Adam's method with default parameters (see, e.g.,
               https://pytorch.org/docs/stable/optim.html);
            -  loss: self.loss (if none, the Mean Squared Loss is set by
               default).
            -  stopping criterion: the fulfillment of the requested tolerance on
               the training loss compatibly with the prescribed budget of
               training iterations (if type(self.stop_training) is list); if
               type(self.stop_training) is int or type(self.stop_training) is
               float, only the number of maximum iterations or the accuracy
               level on the training loss is considered as the stopping rule,
               respectively.

        :param numpy.ndarray points: the coordinates of the given (training)
            points.
        :param numpy.ndarray values: the (training) values in the points.
        """

        self._build_model(points, values)
        self.optimizer = torch.optim.Adam(self.model.parameters())

        points = self._convert_numpy_to_torch(points)
        values = self._convert_numpy_to_torch(values)

        n_epoch = 1
        flag = True
        while flag:
            y_pred = self.model(points)
            loss = self.loss(y_pred, values)
            self.optimizer.zero_grad()
            loss.backward()
            self.optimizer.step()
            self.loss_trend.append(loss.item())
            for criteria in self.stop_training:
                if isinstance(criteria, int):  # stop criteria is an integer
                    if n_epoch == criteria:
                        flag = False
                elif isinstance(criteria, float):  # stop criteria is float
                    if loss.item() < criteria:
                        flag = False

            n_epoch += 1

    @task(returns=np.ndarray, target_direction=IN)
    def predict(self, new_point, scaler_red):
        """
        Evaluate the ANN at given 'new_points'.

        :param array_like new_points: the coordinates of the given points.
        :return: the predicted values via the ANN.
        :rtype: numpy.ndarray
        """
        new_point = self._convert_numpy_to_torch(np.array(new_point))
        y_new = self.model(new_point)
        predicted_red_sol = np.atleast_2d(self._convert_torch_to_numpy(y_new))
        if scaler_red:  # rescale modal coefficients
            predicted_red_sol = scaler_red.inverse_transform(
                predicted_red_sol)
        predicted_red_sol = predicted_red_sol.T
        return predicted_red_sol


import numpy as np
import warnings

from unittest import TestCase
from ezyrb import RadiusNeighborsRegressor, POD, Database, ReducedOrderModel

class TestRadius(TestCase):
    def test_params(self):
        reg = RadiusNeighborsRegressor(radius=3.0, algorithm='kd_tree')
        assert reg.regressor.get_params()['radius'] == 3.0
        assert reg.regressor.get_params()['algorithm'] == 'kd_tree'

    def test_fit_onescalarparam_scalarfunc(self):
        reg = RadiusNeighborsRegressor()
        reg.fit([1], [20])
        assert reg.regressor.n_samples_fit_ == 1

    def test_fit_scalarparam_scalarfunc(self):
        reg = RadiusNeighborsRegressor()
        reg.fit([1, 2, 5, 7, 2], [2, 5, 7, 83, 3])
        assert reg.regressor.n_samples_fit_ == 5

    def test_fit_biparam_scalarfunc(self):
        reg = RadiusNeighborsRegressor()
        reg.fit([[1, 2], [6, 7], [8, 9]], [1, 5, 6])
        assert reg.regressor.n_samples_fit_ == 3

    def test_fit_biparam_bifunc(self):
        reg = RadiusNeighborsRegressor()
        reg.fit([[1, 2], [6, 7], [8, 9]], [[1, 0], [20, 5], [8, 6]])
        assert reg.regressor.n_samples_fit_ == 3

    def test_radiusneighbors(self):
        reg = RadiusNeighborsRegressor(radius=3.0)
        reg.fit([[1, 2], [6, 7], [8, 9]], [[1, 0], [20, 5], [8, 6]])
        neigh_idx = reg.regressor.radius_neighbors([[7,8]], return_distance=False)[0]
        assert neigh_idx[0] == 1
        assert neigh_idx[1] == 2
        assert len(neigh_idx) == 2

    def test_predict(self):
        reg = RadiusNeighborsRegressor(radius=0.5)
        reg.fit([[1, 2], [6, 7], [8, 9]], [[1, 0], [20, 5], [8, 6]])
        neigh_idx = reg.regressor.predict([[1,2], [8,9], [6,7]])
        assert (neigh_idx[0] == [1,0]).all()
        assert (neigh_idx[1] == [8,6]).all()
        assert (neigh_idx[2] == [20, 5]).all()

    def test_with_db_predict(self):
        reg = RadiusNeighborsRegressor(radius=0.5)
        pod = POD()
        db = Database(np.array([1, 2, 3])[:,None], np.array([1, 5, 3])[:,None])
        rom = ReducedOrderModel(db, pod, reg)

        rom.fit()
        assert rom.predict([1]) == 1
        assert rom.predict([2]) == 5
        assert rom.predict([3]) == 3

    def test_wrong1(self):
        # wrong number of params
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", category=np.VisibleDeprecationWarning) 
            with self.assertRaises(Exception):
                reg = RadiusNeighborsRegressor()
                reg.fit([[1, 2], [6,], [8, 9]], [[1, 0], [20, 5], [8, 6]])

    def test_wrong2(self):
        # wrong number of values
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", category=np.VisibleDeprecationWarning) 
            with self.assertRaises(Exception):
                reg = RadiusNeighborsRegressor()
                reg.fit([[1, 2], [6,], [8, 9]], [[20, 5], [8, 6]])



"""
Module for FNN-Autoencoders.
"""

import sys      # Path to find PyEDDL library in PyCOMPSs container
sys.path.append('/usr/local/miniconda3/lib/python3.8/site-packages')
from pyeddl import eddl
from pyeddl.tensor import Tensor
import numpy as np
from pycompss.api.task import task
from pycompss.api.parameter import INOUT, IN
from .reduction import Reduction

class AE_EDDL(Reduction):
    """
    Feed-Forward AutoEncoder class (AE)

    :param list layers_encoder: ordered list with the number of neurons of
        each hidden layer for the encoder
    :param list layers_decoder: ordered list with the number of neurons of
        each hidden layer for the decoder
    :param pyeddl.eddl.activation function_encoder: activation function
        at each layer for the encoder, except for the output layer at with
        Identity is considered by default.  A single activation function can
        be passed or a list of them of length equal to the number of hidden
        layers.
    :param pyeddl.eddl.activation function_decoder: activation function
        at each layer for the decoder, except for the output layer at with
        Identity is considered by default.  A single activation function can
        be passed or a list of them of length equal to the number of hidden
        layers.
    :param list stop_training: list with the maximum number of training
        iterations (int) and/or the desired tolerance on the training loss
        (float).
    :param int pyeddl.eddl batch_size: size of data batches used for
        training the network.
    :param pyeddl.eddl optimizer: the eddl class implementing optimizer
        Default value is `Adam` optimizer.
    :param str pyeddl.eddl loss: loss definition (Mean Squared if not
        given).
    :param str pyeddl.eddl metric: metric definition (Mean Squared if not
        given).
    :param float lr: the learning rate. Default is 0.001.
    :param cs: type and number of the computing service. Default is
        eddl.CS_CPU().
    :param int training_type: 1 for Coarse training, 2 Fine training_1
        3 for Fine training_2. Default is 2.

    :Example:
        >>> from ezyrb import AE_EDDL
        >>> import pyeddl.eddl as eddl
        >>> from pyeddl.tensor import Tensor
        >>> f = eddl.Tanh
        >>> low_dim = 5
        >>> optim = eddl.adam
        >>> ae = AE_EDDL([400, low_dim], [low_dim, 400], f(), f(), 2000)
        >>> # or ...
        >>> ae = AE_EDDL([400, 10, 10, low_dim], [low_dim, 400], f(), f(), 1e-5,
        >>>          optimizer=optim)
        >>> ae.fit(snapshots)
        >>> reduced_snapshots = ae.reduce(snapshots)
        >>> expanded_snapshots = ae.expand(reduced_snapshots)
    """
    def __init__(self,
                 layers_encoder,
                 layers_decoder,
                 function_encoder,
                 function_decoder,
                 stop_training,
                 batch_size,
                 loss=None,
                 metric=None,
                 optimizer=eddl.adam,
                 lr=0.001,
                 cs = eddl.CS_CPU,
                 training_type = 2):

        if loss is None:
            loss = "mean_squared_error"
        if metric is None:
            metric = "mean_squared_error"

        if not isinstance(function_encoder, list):
            # Single activation function passed
            function_encoder = [function_encoder] * (len(layers_encoder))
        if not isinstance(function_decoder, list):
            # Single activation function passed
            function_decoder = [function_decoder] * (len(layers_decoder))

        if not isinstance(stop_training, list):
            stop_training = [stop_training]

        self.layers_encoder = layers_encoder
        self.layers_decoder = layers_decoder
        self.function_encoder = function_encoder
        self.function_decoder = function_decoder
        self.loss = loss
        self.metric = metric

        self.stop_training = stop_training
        self.batch_size = batch_size
        self.optimizer = optimizer
        self.lr = lr
        self.cs = cs
        self.training_type = training_type
        self.loss_trend = []
        self.metric_trend = []
        self.encoder = None
        self.decoder = None
        self.decoder2 = None
        self.model_Autoencoder = None
        self.model_Decoder = None
        self.file_1 = "trained_model_Autoencoder.onnx"
        self.file_2 = "trained_model_Decoder.onnx"
        self.fitted = False
        self.imported = False

    def __getstate__(self):
        """
        Used for serializing instances.

        Will be invoked only when using PyCOMPSs:
            -  objects used as task parameters must be automatically
               serializable/picklable.
            -  So, we save the trained model and delete all unpicklables at the
               time of serialization.
        """

        # Start with a copy so as not to accidentally modify the object state or
        # cause other conflicts
        state = self.__dict__.copy()

        # remove unpicklable entries
        del state['encoder']
        del state['decoder']
        del state['decoder2']
        del state['model_Autoencoder']
        del state['model_Decoder']
        return state

    def __setstate__(self, state):
        """
        Used for deserializing.

        Will be invoked only when using PyCOMPSs:
            -  objects used as task parameters must be automatically
               serializable/picklable.
            -  So, we save the trained model and delete all unpicklables at the
               time of serialization.
        """

        # Restore the state which was picklable and add the picklable to the
        # state (happen at the end/beginning of eache pycompss tasks)
        self.__dict__.update(state)

        # restore unpicklable entries
        if self.fitted: # inputs for all tasks except the fit() task
            self.imported = True

            # (n) hidden layers + (n-1) activation layers + (1) input layer
            n = len(self.layers_encoder)
            encoder_layer_index = 2*n - 1
            self.model_Autoencoder = eddl.import_net_from_onnx_file(self.file_1)
            self.encoder = self.model_Autoencoder.layers[encoder_layer_index]
            self.decoder = self.model_Autoencoder.layers[-1]

            eddl.build(
                self.model_Autoencoder,
                self.optimizer(self.lr),
                [self.loss ],
                [self.metric],
                self.cs(mem="low_mem"),
                False # don't initialize random weights (using trained model)
            )
            # resize manually since we don't use "fit" -->
            # size of each layer = (batch_size, layer_dof)
            self.model_Autoencoder.resize(self.batch_size)
            #-------------------------------------------------------------------
            self.model_Decoder = eddl.import_net_from_onnx_file(self.file_2)
            self.decoder2 = self.model_Decoder.layers[-1]

            eddl.build(
                self.model_Decoder,
                self.optimizer(self.lr),
                [self.loss ],
                [self.metric],
                self.cs(mem="low_mem"),
                False # don't initialize random weights (using trained model)
            )
            # resize manually since we don't use "fit" -->
            # size of each layer = (batch_size, layer_dof)
            self.model_Decoder.resize(self.batch_size)

        else: # inputs for the fit() method
            self.encoder = None
            self.decoder = None
            self.decoder2 = None
            self.model_Autoencoder = None
            self.model_Decoder = None

    def _build_model(self, values):
        """
        Build the PyEDDL model.

        Considering the number of neurons per layer (self.layers), a
        feed-forward NN is defined:
            - activation function from layer i>=0 to layer i+1:
              self.function[i]; activation function at the output layer:
              Identity (by default).

        :param numpy.ndarray values: the set values one wants to reduce.
        """
        layers_encoder = self.layers_encoder.copy()
        layers_encoder.insert(0, values.shape[1])

        layers_decoder = self.layers_decoder.copy()
        layers_decoder.append(values.shape[1])

        def EncoderBlock(layer):
            for i in range(1, len(layers_encoder)-1):
                layer = self.function_encoder[i](
                    eddl.Dense(layer, layers_encoder[i]))
            layer = eddl.Dense(layer, layers_encoder[-1])
            return layer

        def DecoderBlock(layer):
            for i in range(1, len(layers_decoder)-1):
                layer = self.function_encoder[i](
                    eddl.Dense(layer, layers_decoder[i]))
            layer = eddl.Dense(layer, layers_decoder[-1])
            return layer    
        #-----------------------------------------------------------------------
        # Define network1
        in1 = eddl.Input([layers_encoder[0]])
        self.encoder = EncoderBlock(in1)
        self.decoder = DecoderBlock(self.encoder)
        self.model_Autoencoder = eddl.Model([in1], [self.decoder])

        eddl.build(
            self.model_Autoencoder,
            self.optimizer(self.lr),
            [self.loss ],
            [self.metric],
            self.cs(mem="low_mem"),
            True # initialize weights to random values
        )
        #-----------------------------------------------------------------------
        # Define network2
        in2 = eddl.Input([layers_decoder[0]])
        self.decoder2 = DecoderBlock(in2)
        self.model_Decoder = eddl.Model([in2], [self.decoder2])
    
        eddl.build(
            self.model_Decoder,
            self.optimizer(self.lr),
            [self.loss ],
            [self.metric],
            self.cs(mem="low_mem"),
            True # initialize weights to random values
        )
        #-----------------------------------------------------------------------
        eddl.summary(self.model_Autoencoder)
        eddl.plot(self.model_Autoencoder, "Autoencoder.pdf")
        eddl.summary(self.model_Decoder)
        eddl.plot(self.model_Decoder, "Decoder.pdf")
        #-----------------------------------------------------------------------

    @task(target_direction=INOUT)
    def fit(self, values):
        """
        Build the AE given 'values' and perform training.

        Training procedure information:
            -  optimizer: Adam's method with default parameters.
            -  loss: self.loss (if none, the Mean Squared Loss is set by
               default).
            -  stopping criterion: the fulfillment of the requested tolerance
               on the training loss compatibly with the prescribed budget of
               training iterations (if type(self.stop_training) is list); if
               type(self.stop_training) is int or type(self.stop_training) is
               float, only the number of maximum iterations or the accuracy
               level on the training loss is considered as the stopping rule,
               respectively.

        :param numpy.ndarray values: the (training) values in the points.
        """
        values = values.T
        values = Tensor.fromarray(values) # Numpy array to EDDL.Tensor
        self._build_model(values)
        #-----------------------------------------------------------------------
        if self.training_type == 1:
            print('Coarse training:')
            n_epoch =1
            flag = True
            while flag:
                print('Iteration number: ',n_epoch)
                eddl.fit(self.model_Autoencoder, [values], [values],
                    self.batch_size, 1)
                losses = eddl.get_losses(self.model_Autoencoder)
                metrics = eddl.get_metrics(self.model_Autoencoder)
                self.loss_trend.append(losses)
                self.metric_trend.append(metrics)

                for criteria in self.stop_training:
                    if isinstance(criteria, int):
                        # stop criteria is an integer
                        if n_epoch == criteria:
                            flag = False
                    elif isinstance(criteria, float):
                        # stop criteria is float
                        if losses[0] < criteria:
                            flag = False
                n_epoch += 1
        #-----------------------------------------------------------------------
        elif self.training_type == 2:
            print('Fine training_1:')
            s = values.shape
            num_batches = s[0] // self.batch_size
            xbatch = Tensor([self.batch_size, 3067])
            n_epoch =1
            flag = True
            while flag:
                eddl.reset_loss(self.model_Autoencoder)
                for j in range(num_batches):
                    # 1) using next_batch
                    eddl.next_batch([values], [xbatch])
                    eddl.train_batch(self.model_Autoencoder, [xbatch], [xbatch])

                losses = eddl.get_losses(self.model_Autoencoder)
                metrics = eddl.get_metrics(self.model_Autoencoder)
                self.loss_trend.append(losses)
                self.metric_trend.append(metrics)

                ## prints are turned off in the fine training
                # print("Epoch %d (%d batches)" % (n_epoch, num_batches))
                # for l, m in zip(losses, metrics):
                #     print("Loss: %.6f\tMetric: %.6f" % (l, m))
                
                for criteria in self.stop_training:
                    if isinstance(criteria, int):
                        # stop criteria is an integer
                        if n_epoch == criteria:
                            flag = False
                    elif isinstance(criteria, float):
                        # stop criteria is float
                        if losses[0] < criteria:
                            flag = False
                n_epoch += 1
        #-----------------------------------------------------------------------
        elif self.training_type == 3:
            print('Fine training_2:')
            s = values.shape
            num_batches = s[0] // self.batch_size
            n_epoch =1
            flag = True
            while flag:
                eddl.reset_loss(self.model_Autoencoder)
                for j in range(num_batches):
                    # 2) using samples indices
                    indices = np.random.randint(0, s[0], self.batch_size)
                    eddl.train_batch(self.model_Autoencoder, [values], [values],
                        indices)

                losses = eddl.get_losses(self.model_Autoencoder)
                metrics = eddl.get_metrics(self.model_Autoencoder)
                self.loss_trend.append(losses)
                self.metric_trend.append(metrics)

                ## prints are turned off in the fine training
                # print("Epoch {} ({} batches)".format(n_epoch, num_batches))
                # for l, m in zip(losses, metrics):
                #     print("Loss: %.6f\tMetric: %.6f" % (l, m))
                
                for criteria in self.stop_training:
                    if isinstance(criteria, int):
                        # stop criteria is an integer
                        if n_epoch == criteria:
                            flag = False
                    elif isinstance(criteria, float):
                        # stop criteria is float
                        if losses[0] < criteria:
                            flag = False
                n_epoch += 1
        #-----------------------------------------------------------------------
        # Copy parameters from model_Autoencoder to model_Decoder
        # munber of Decoder layers = (n) hidden layers + (n-1) activation layers
        n = len(self.layers_encoder)
        num_lay_decoder = 2*n - 1
        decoder_parameters = eddl.get_parameters(self.model_Autoencoder,
            True)[-num_lay_decoder:]
        # Insert empty parameter for the new input layer of decoder
        decoder_parameters.insert(0,[])
        eddl.set_parameters(self.model_Decoder, decoder_parameters)
        ## For debugging
        # for i in decoder_parameters:
        #     print(len(i))
        #-----------------------------------------------------------------------
        # For PyCOMPSs: objects used as task parameters must be automatically
        # serializable/picklable so we save the trained model and delete all
        # unpicklables at the time of serialization
        eddl.save_net_to_onnx_file(self.model_Autoencoder, self.file_1)
        eddl.save_net_to_onnx_file(self.model_Decoder, self.file_2)
        self.fitted = True

    @task(returns=np.ndarray, target_direction=IN)
    def transform(self, X, scaler_red):
        """
        Reduces the given snapshots.

        :param numpy.ndarray X: the input snapshots matrix (stored by column).
        """
        #-----------------------------------------------------------------------
        if self.imported: # Means PyCOMPSs is used.
            # # For debugging
            # print("Encoder layer {} --> {}".format(
            #     self.encoder.input.shape, self.encoder.output.shape))
            print(f"Trained model imported from ({self.file_1})")
            eddl.summary(self.model_Autoencoder)
        #-----------------------------------------------------------------------
        X = Tensor.fromarray(X.T) # Numpy array to EDDL.Tensor
        # One prediction for the fitted model(1 forward pass after training)
        eddl.predict(self.model_Autoencoder, [X])
        g = eddl.getOutput(self.encoder)
        reduced_output = (Tensor.getdata(g).T).T # EDDL.Tensor to Numpy array
        if scaler_red:
            reduced_output = scaler_red.fit_transform(reduced_output)

        ## For debugging
        # u = eddl.getOutput(self.decoder)
        # print('Latent sapce info.:')
        # g.info()
        # print('Expansion sapce info.:')
        # u.info()

        return reduced_output

    @task(returns=np.ndarray, target_direction=IN)
    def inverse_transform(self, g, database):
        """
        Projects a reduced to full order solution.

        :param: numpy.ndarray g the latent variables.
        """
        #-----------------------------------------------------------------------
        if self.imported: # Means PyCOMPSs is used.
            print(f"Trained model imported from ({self.file_2})")
            eddl.summary(self.model_Decoder)
        #-----------------------------------------------------------------------
        g = Tensor.fromarray(g.T) # Numpy array to EDDL.Tensor
        # One forward pass for the new decoder (without training, parameters
        # copied from the trained autoencoder)
        eddl.forward(self.model_Decoder, [g])
        u = eddl.getOutput(self.decoder2)

        ## For debugging
        # print('Before expansion', g.shape)
        # print('Before expansion', u.shape)

        predicted_sol = Tensor.getdata(u).T # EDDL.Tensor to Numpy array
        
        if database and database.scaler_snapshots:
            predicted_sol = database.scaler_snapshots.inverse_transform(
                    predicted_sol.T).T

        if 1 in predicted_sol.shape:
            predicted_sol = predicted_sol.ravel()
        else:
            predicted_sol = predicted_sol.T
        
        return predicted_sol

    def reduce(self, X, scaler_red):
        """
        Reduces the given snapshots.

        :param numpy.ndarray X: the input snapshots matrix (stored by column).

        .. note::

            Same as `transform`. Kept for backward compatibility.
        """
        return self.transform(X, scaler_red)

    def expand(self, g, database):
        """
        Projects a reduced to full order solution.

        :param: numpy.ndarray g the latent variables.

        .. note::

            Same as `inverse_transform`. Kept for backward compatibility.
        """
        return self.inverse_transform(g, database)



import numpy as np

from unittest import TestCase
from ezyrb import Database


class TestDatabase(TestCase):
    def test_constructor_empty(self):
        a = Database()

    def test_constructor_arg(self):
        Database(np.random.uniform(size=(10, 3)),
                 np.random.uniform(size=(10, 8)))

        Database(np.random.uniform(size=(10, 3)),
                 np.random.uniform(size=(10, 8)),
                 np.random.uniform(size=(10, 8)),)

    def test_constructor_arg_wrong(self):
        with self.assertRaises(RuntimeError):
            Database(np.random.uniform(size=(9, 3)),
                     np.random.uniform(size=(10, 8)))

    def test_constructor_arg_wrong_space_len(self):  
        with self.assertRaises(RuntimeError):
            Database(np.random.uniform(size=(10, 3)),
                     np.random.uniform(size=(10, 9)),
                     space = np.random.uniform(size=(9,9)))
    
    def test_constructor_arg_wrong_space_width(self):  
        with self.assertRaises(RuntimeError):
            Database(np.random.uniform(size=(10, 3)),
                     np.random.uniform(size=(10, 9)),
                     space = np.random.uniform(size=(10,8)))

    def test_constructor_error(self):
        with self.assertRaises(RuntimeError):
            Database(np.eye(5))

    def test_getitem(self):
        org = Database(np.random.uniform(size=(10, 3)),
                       np.random.uniform(size=(10, 8)))
        new = org[2::2]
        assert new.parameters.shape[0] == new.snapshots.shape[0] == 4
    
    def test_getitem_singular(self): 
        org = Database(np.random.uniform(size=(10, 3)),
                       np.random.uniform(size=(10, 8)))
        new = org[2]
        assert True
    
    def test_getitem_space(self):
        org = Database(np.random.uniform(size=(10, 3)),
                       np.random.uniform(size=(10, 8)),
                       space = np.random.uniform(size=(10,8)))
        new = org[2::2]
        assert new.parameters.shape[0] == new.snapshots.shape[0] == new.space.shape[0] == 4
    




import numpy as np

from unittest import TestCase
from ezyrb import GPR
import sklearn

np.random.seed(17)

def get_xy():
    npts = 20
    dinput = 4

    inp = np.random.uniform(-1, 1, size=(npts, dinput))
    out = np.array([
        np.sin(inp[:, 0]) + np.sin(inp[:, 1]**2),
        np.cos(inp[:, 2]) + np.cos(inp[:, 3]**2)
    ]).T

    return inp, out

class TestGPR(TestCase):
    def test_constructor_empty(self):
        gpr = GPR()

    def test_types(self):
        x, y = get_xy()
        gpr = GPR()
        gpr.fit(x[:, 0], y[:, 0])
        assert isinstance(gpr.model,
                          sklearn.gaussian_process.GaussianProcessRegressor)

    def test_predict_01(self):
        x, y = get_xy()
        gpr = GPR()
        gpr.fit(x, y, optimization_restart=50)
        test_y, variance = gpr.predict(x, return_variance=True)
        np.testing.assert_array_almost_equal(y, test_y, decimal=6)

    def test_predict_02(self):
        np.random.seed(42)
        x, y = get_xy()
        gpr = GPR()
        gpr.fit(x, y, optimization_restart=50)
        test_y, variance = gpr.predict(x[:4], return_variance=True)
        true_var = np.array([[5.761689e-06, 2.017326e-06],
                             [5.761686e-06, 2.017325e-06],
                             [5.761692e-06, 2.017327e-06],
                             [5.761695e-06, 2.017328e-06]])
        np.testing.assert_array_almost_equal(true_var, variance, decimal=6)

    def test_predict_03(self):
        np.random.seed(1)
        x, y = get_xy()
        gpr = GPR()
        gpr.fit(x, y, optimization_restart=50)
        test_y = gpr.predict(x)
        np.testing.assert_array_almost_equal(y, test_y, decimal=6)

    def test_optimal_mu(self):
        x, y = get_xy()
        gpr = GPR()
        gpr.fit(x, y, optimization_restart=10)
        new_mu = gpr.optimal_mu(np.array([[-1, 1], [-1, 1], [-1, 1], [-1, 1]]))
        np.testing.assert_array_almost_equal(np.abs(new_mu), np.ones(4).reshape(1, -1))



import numpy as np
import torch

from unittest import TestCase
from ezyrb import AE, POD, PODAE

snapshots = np.load('tests/test_datasets/p_snapshots.npy').T


class TestAE(TestCase):
    def test_constructor_empty(self):
        ae = AE([20, 2], [2, 20], torch.nn.ReLU, torch.nn.ReLU, 20)
        pod = POD()
        PODAE(pod, ae)

    def test_reconstruction(self):
        f = torch.nn.Softplus
        ae = AE([20, 20, 2], [2, 20, 20], f(), f(), 1e-9)
        pod = POD(rank=4)
        podae = PODAE(pod, ae)
        podae.fit(snapshots)
        snapshots_ = podae.inverse_transform(podae.transform(snapshots))
        rerr = np.linalg.norm(snapshots_ - snapshots)/np.linalg.norm(snapshots)
        assert rerr < 5e-3

    #TODO
    # def test_decode_encode(self):
    #     f = torch.nn.Softplus
    #     low_dim = 5
    #     ae = AE([400, low_dim], [low_dim, 400], f(), f(), 20)
    #     ae.fit(snapshots)
    #     reduced_snapshots = ae.reduce(snapshots)
    #     assert reduced_snapshots.shape[0] == low_dim
    #     expanded_snapshots = ae.expand(reduced_snapshots)
    #     assert expanded_snapshots.shape[0] == snapshots.shape[0]



import numpy as np

from unittest import TestCase
from ezyrb import POD, GPR, RBF, Database
from ezyrb import KNeighborsRegressor, RadiusNeighborsRegressor, Linear
from ezyrb import ReducedOrderModel as ROM

snapshots = np.load('tests/test_datasets/p_snapshots.npy').T
pred_sol_tst = np.load('tests/test_datasets/p_predsol.npy').T
pred_sol_gpr = np.load('tests/test_datasets/p_predsol_gpr.npy').T
param = np.array([[-.5, -.5], [.5, -.5], [.5, .5], [-.5, .5]])


class TestReducedOrderModel(TestCase):
    def test_constructor(self):
        pod = POD()
        rbf = RBF()
        db = Database(param, snapshots.T)
        rom = ROM(db, pod, rbf)

    def test_save(self):
        fname = 'ezyrb.tmp'
        pod = POD()
        rbf = RBF()
        db = Database(param, snapshots.T)
        rom = ROM(db, pod, rbf)
        rom.fit()
        rom.save(fname)

    def test_load(self):
        fname = 'ezyrb.tmp'
        pod = POD()
        rbf = RBF()
        db = Database(param, snapshots.T)
        rom = ROM(db, pod, rbf)
        rom.fit()
        rom.save(fname)
        new_rom = ROM.load(fname)
        new_param = [-0.293344, -0.23120537]
        np.testing.assert_array_almost_equal(
            rom.predict(new_param),
            new_rom.predict(new_param)
        )

    def test_load2(self):
        fname = 'ezyrb2.tmp'
        pod = POD()
        rbf = RBF()
        db = Database(param, snapshots.T)
        rom = ROM(db, pod, rbf)
        rom.fit()
        rom.save(fname, save_db=False)
        new_rom = ROM.load(fname)
        new_param = [-0.293344, -0.23120537]
        np.testing.assert_array_almost_equal(
            rom.predict(new_param),
            new_rom.predict(new_param)
        )

    def test_predict_01(self):
        pod = POD()
        rbf = RBF()
        db = Database(param, snapshots.T)
        rom = ROM(db, pod, rbf).fit()
        pred_sol = rom.predict([-0.293344, -0.23120537])
        np.save('tests/test_datasets/p_predsol.npy', pred_sol.T)
        np.testing.assert_allclose(pred_sol, pred_sol_tst, rtol=1e-4, atol=1e-5)

    def test_predict_02(self):
        np.random.seed(117)
        pod = POD(method='svd', rank=4)
        gpr = GPR()
        db = Database(param, snapshots.T)
        rom = ROM(db, pod, gpr).fit()
        pred_sol = rom.predict([-.45, -.45])
        np.testing.assert_allclose(pred_sol, pred_sol_gpr, rtol=1e-4, atol=1e-5)

    def test_predict_03(self):
        pod = POD(method='svd', rank=3)
        gpr = GPR()
        db = Database(param, snapshots.T)
        rom = ROM(db, pod, gpr).fit()
        pred_sol = rom.predict(db.parameters[2])
        assert pred_sol.shape == db.snapshots[0].shape

    def test_predict_04(self):
        pod = POD(method='svd', rank=3)
        gpr = GPR()
        db = Database(param, snapshots.T)
        rom = ROM(db, pod, gpr).fit()
        pred_sol = rom.predict(db.parameters)
        assert pred_sol.shape == db.snapshots.shape

    def test_predict_scaler_01(self):
        from sklearn.preprocessing import StandardScaler
        scaler = StandardScaler()
        pod = POD()
        rbf = RBF()
        db = Database(param, snapshots.T, scaler_snapshots=scaler)
        rom = ROM(db, pod, rbf).fit()
        pred_sol = rom.predict(db.parameters[0])
        np.testing.assert_allclose(pred_sol, db._snapshots[0], rtol=1e-4, atol=1e-5)
        pred_sol = rom.predict(db.parameters[0:2])
        np.testing.assert_allclose(pred_sol, db._snapshots[0:2], rtol=1e-4, atol=1e-5)

    def test_predict_scaler_02(self):
        from sklearn.preprocessing import StandardScaler
        scaler_p = StandardScaler()
        scaler_s = StandardScaler()
        pod = POD()
        rbf = RBF()
        db = Database(param, snapshots.T, scaler_parameters=scaler_p, scaler_snapshots=scaler_s)
        rom = ROM(db, pod, rbf).fit()
        pred_sol = rom.predict(db._parameters[0])
        np.testing.assert_allclose(pred_sol, db._snapshots[0], rtol=1e-4, atol=1e-5)
        pred_sol = rom.predict(db._parameters[0:2])
        np.testing.assert_allclose(pred_sol, db._snapshots[0:2], rtol=1e-4, atol=1e-5)

    def test_predict_scaling_coeffs(self):
        from sklearn.preprocessing import StandardScaler
        scaler = StandardScaler()
        pod = POD()
        rbf = RBF()
        db = Database(param, snapshots.T)
        rom = ROM(db, pod, rbf, scaler).fit()
        pred_sol = rom.predict(db._parameters[0])
        np.testing.assert_allclose(pred_sol, db._snapshots[0], rtol=1e-4, atol=1e-5)
        pred_sol = rom.predict(db._parameters[0:2])
        np.testing.assert_allclose(pred_sol, db._snapshots[0:2], rtol=1e-4, atol=1e-5)

    def test_test_error(self):
        pod = POD(method='svd', rank=-1)
        rbf = RBF()
        db = Database(param, snapshots.T)
        rom = ROM(db, pod, rbf).fit()
        error = rom.test_error(db)
        np.testing.assert_almost_equal(error, 0, decimal=6)

    def test_kfold_cv_error_01(self):
        pod = POD()
        rbf = RBF()
        db = Database(param, snapshots.T)
        n_splits = len(db)
        rom = ROM(db, pod, rbf)
        err_kfold = rom.kfold_cv_error(n_splits=n_splits)
        err_loo = rom.loo_error()
        np.testing.assert_allclose(err_kfold, err_loo)

    def test_loo_error_01(self):
        pod = POD()
        rbf = RBF()
        gpr = GPR()
        rnr = RadiusNeighborsRegressor()
        knr = KNeighborsRegressor(n_neighbors=1)
        lin = Linear()
        db = Database(param, snapshots.T)
        exact_len = len(db)
        approximations = [rbf, gpr, knr, rnr, lin]
        roms = [ROM(db, pod, app) for app in approximations]
        len_errors = [len(rom.loo_error()) for rom in roms]
        np.testing.assert_allclose(len_errors, exact_len)

    def test_loo_error_02(self):
        pod = POD()
        gpr = GPR()
        db = Database(param, snapshots.T)
        rom = ROM(db, pod, gpr)
        err = rom.loo_error(normalizer=False)
        np.testing.assert_allclose(
            err[0],
            np.array(0.639247),
            rtol=1e-3)

    def test_loo_error_singular_values(self):
        pod = POD()
        rbf = RBF()
        db = Database(param, snapshots.T)
        rom = ROM(db, pod, rbf).fit()
        valid_svalues = rom.reduction.singular_values
        rom.loo_error()
        np.testing.assert_allclose(valid_svalues, rom.reduction.singular_values)

    def test_optimal_mu(self):
        pod = POD()
        rbf = RBF()
        gpr = GPR()
        rnr = RadiusNeighborsRegressor()
        knr = KNeighborsRegressor(n_neighbors=1)
        lin = Linear()
        db = Database(param, snapshots.T)
        exact_len = param.shape[1]
        approximations = [rbf, gpr, knr, rnr, lin]
        for k in [1, 2]:
            roms = [ROM(db, pod, app).fit() for app in approximations]
            len_opt_mu = [rom.optimal_mu(k=k).shape[1] for rom in roms]
            np.testing.assert_allclose(len_opt_mu, exact_len)
            len_k = [rom.optimal_mu(k=k).shape[0] for rom in roms]
            np.testing.assert_allclose(len_k, k)




"""
Module for Linear N-Dimensional Interpolation
"""

import numpy as np
from scipy.interpolate import LinearNDInterpolator as LinearNDInterp
from scipy.interpolate import interp1d

from pycompss.api.task import task
from pycompss.api.parameter import INOUT, IN
from .approximation import Approximation

class Linear(Approximation):
    """
    Multidimensional linear interpolator.

    :param float fill_value: value used to fill in for requested points outside
        of the convex hull of the input points. If not provided, then the
        default is numpy.nan.
    """
    def __init__(self, fill_value=np.nan):
        self.fill_value = fill_value
        self.interpolator = None

    @task(target_direction=INOUT)
    def fit(self, points, values):
        """
        Construct the interpolator given `points` and `values`.

        :param array_like points: the coordinates of the points.
        :param array_like values: the values in the points.
        """
        # the first dimension is the list of parameters, the second one is
        # the dimensionality of each tuple of parameters (we look for
        # parameters of dimensionality one)
        as_np_array = np.array(points)
        if not np.issubdtype(as_np_array.dtype, np.number):
            raise ValueError('Invalid format or dimension for the argument'
                             '`points`.')

        if as_np_array.shape[-1] == 1:
            as_np_array = np.squeeze(as_np_array, axis=-1)

        if as_np_array.ndim == 1 or (as_np_array.ndim == 2
                                     and as_np_array.shape[1] == 1):
            self.interpolator = interp1d(as_np_array, values, axis=0)
        else:
            self.interpolator = LinearNDInterp(points,
                                               values,
                                               fill_value=self.fill_value)

    @task(returns=np.ndarray, target_direction=IN)
    def predict(self, new_point, scaler_red):
        """
        Evaluate interpolator at given `new_points`.

        :param array_like new_points: the coordinates of the given points.
        :return: the interpolated values.
        :rtype: numpy.ndarray
        """
        new_red_snap = self.interpolator(new_point)

        # Drop the fist 1 from resulted from interp1d
        # (1, 1, latent) --> (1, latent)
        if (new_red_snap.shape[0] == 1) and (new_red_snap.shape[1] == 1):
            new_red_snap = np.squeeze(new_red_snap, axis=0)

        predicted_red_sol = np.atleast_2d(new_red_snap)

        if scaler_red:  # rescale modal coefficients
            predicted_red_sol = scaler_red.inverse_transform(
                predicted_red_sol)
        predicted_red_sol = predicted_red_sol.T

        return predicted_red_sol



import numpy as np
import torch.nn as nn
from torch import Tensor, from_numpy

from unittest import TestCase
from ezyrb import ANN

np.random.seed(17)

def get_xy():
    npts = 20
    dinput = 4

    inp = np.random.uniform(-1, 1, size=(npts, dinput))
    out = np.array([
        np.sin(inp[:, 0]) + np.sin(inp[:, 1]**2),
        np.cos(inp[:, 2]) + np.cos(inp[:, 3]**2)
    ]).T

    return inp, out

class TestANN(TestCase):
    def test_constructor_empty(self):
        ann = ANN([10, 5], nn.Tanh(), 20000)

    def test_constrctor_loss_none(self):
        ann = ANN([10, 5], nn.Tanh(), 20000, loss=None)
        assert isinstance(ann.loss, nn.MSELoss)

    def test_constructor_single_function(self):
        passed_func = nn.Tanh()
        ann = ANN([10, 5], passed_func, 20000)

        assert isinstance(ann.function, list)
        for func in ann.function:
            assert func == passed_func

    def test_constructor_layers(self):
        ann = ANN([10, 5], nn.Tanh(), 20000)
        assert ann.layers == [10, 5]

    def test_constructor_stop_training(self):
        ann = ANN([10, 5], nn.Tanh(), 20000)
        assert isinstance(ann.stop_training, list)
        assert ann.stop_training == [20000]

    def test_constructor_fields_initialized(self):
        ann = ANN([10, 5], nn.Tanh(), 20000)
        assert ann.loss_trend == []
        assert ann.model is None

    def test_fit_mono(self):
        x, y = get_xy()
        ann = ANN([10, 5], nn.Tanh(), [20000, 1e-5])
        ann.fit(x[:, 0].reshape(-1,1), y[:, 0].reshape(-1,1))
        assert isinstance(ann.model, nn.Sequential)

    def test_fit_01(self):
        x, y = get_xy()
        ann = ANN([10, 5], nn.Tanh(), [20000, 1e-8])
        ann.fit(x, y)
        assert isinstance(ann.model, nn.Sequential)

    def test_fit_02(self):
        x, y = get_xy()
        ann = ANN([10, 5, 2], [nn.Tanh(), nn.Sigmoid(), nn.Tanh()], [20000, 1e-8])
        ann.fit(x, y)
        assert isinstance(ann.model, nn.Sequential)

    def test_predict_01(self):
        np.random.seed(1)
        x, y = get_xy()
        ann = ANN([10, 5], nn.Tanh(), 20)
        ann.fit(x, y)
        test_y = ann.predict(x)
        assert isinstance(test_y, np.ndarray)

    def test_predict_02(self):
        np.random.seed(1)
        x, y = get_xy()
        ann = ANN([10, 5], nn.Tanh(), [20000, 1e-8])
        ann.fit(x, y)
        test_y = ann.predict(x)
        np.testing.assert_array_almost_equal(y, test_y, decimal=3)

    def test_predict_03(self):
        np.random.seed(1)
        x, y = get_xy()
        ann = ANN([10, 5], nn.Tanh(), 1e-8)
        ann.fit(x, y)
        test_y = ann.predict(x)
        np.testing.assert_array_almost_equal(y, test_y, decimal=3)

    def test_convert_numpy_to_torch(self):
        arr = [1.0, 2.0, 3.0, 4.0, 5.0]

        ann = ANN([10, 5], nn.Tanh(), 20000)

        value = ann._convert_numpy_to_torch(np.array(arr))
        assert isinstance(value, Tensor)
        for i in range(len(arr)):
            assert value[i] == arr[i]

    def test_convert_torch_to_numpy(self):
        arr = [1.0, 2.0, 3.0, 4.0, 5.0]
        tensor = from_numpy(np.array(arr)).float()

        ann = ANN([10, 5], nn.Tanh(), 20000)

        value = ann._convert_torch_to_numpy(tensor)
        assert isinstance(value, np.ndarray)
        for i in range(len(arr)):
            assert value[i] == arr[i]

    def test_last_linear(self):
        passed_func = nn.Tanh()
        ann = ANN([10, 5, 2], passed_func, 20000, last_identity=True)
        ann._build_model(np.array([[1,2],[3,4]]), np.array([[5,6],[7,8]]))
        last_layer = ann.model[-1]
        assert isinstance(last_layer, nn.Linear)
        assert last_layer.in_features == 2
        assert last_layer.out_features == 2

        passed_func = nn.Tanh()
        ann = ANN([10, 5, 2], passed_func, 20000, last_identity=False)
        ann._build_model(np.array([[1,2],[3,4]]), np.array([[5,6],[7,8]]))
        last_layer = ann.model[-1]
        assert isinstance(last_layer, nn.Tanh)

        passed_func = [nn.Tanh()] * 3 + [nn.ReLU()]
        ann = ANN([10, 5, 2], passed_func, 20000)
        ann._build_model(np.array([[1,2],[3,4]]), np.array([[5,6],[7,8]]))
        last_layer = ann.model[-1]
        assert isinstance(last_layer, nn.ReLU)

    def test_build_model(self):
        passed_func = nn.Tanh()
        ann = ANN([10, 5, 2], passed_func, 20000)

        ann._build_model(np.array([[1,2],[3,4]]), np.array([[5,6],[7,8]]))

        assert len(ann.model) == 6 + 1
        for i in range(7):
            layer = ann.model[i]
            # the last layer, I keep the separated for clarity
            if i == 6:
                assert isinstance(layer, nn.Linear)
            elif i % 2 == 0:
                assert isinstance(layer, nn.Linear)
            else:
                assert layer == passed_func

        # check input and output
        assert ann.model[6].out_features == np.array([[5,6],[7,8]]).shape[1]
        assert ann.model[0].in_features == np.array([[1,2],[3,4]]).shape[1]

        for i in range(0, 5, 2):
            assert ann.model[i].out_features == ann.model[i+2].in_features



import numpy as np

from unittest import TestCase
from ezyrb import POD

snapshots = np.load('tests/test_datasets/p_snapshots.npy').T
poddb = np.load('tests/test_datasets/p_snapshots_pod.npy')
modes = np.load('tests/test_datasets/p_snapshots_pod_modes.npy')


class TestPOD(TestCase):
    def test_constructor_empty(self):
        a = POD()

    def test_numpysvd(self):
        pod = POD('svd').fit(snapshots)
        snapshots_ = pod.inverse_transform(pod.transform(snapshots))
        np.testing.assert_array_almost_equal(snapshots, snapshots_, decimal=4)

    def test_correlation_matrix(self):
        pod = POD('correlation_matrix').fit(snapshots)
        snapshots_ = pod.inverse_transform(pod.transform(snapshots))
        np.testing.assert_array_almost_equal(snapshots, snapshots_, decimal=3)

    def test_correlation_matirix_savemem(self):
        pod = POD('correlation_matrix', save_memory=True).fit(snapshots)
        snapshots_ = pod.inverse_transform(pod.transform(snapshots))
        np.testing.assert_array_almost_equal(snapshots, snapshots_, decimal=3)

    def test_randomized_svd(self):
        pod = POD('randomized_svd', save_memory=False).fit(snapshots)
        snapshots_ = pod.inverse_transform(pod.transform(snapshots))
        np.testing.assert_array_almost_equal(snapshots, snapshots_, decimal=4)

    def test_numpysvd_old(self):
        pod = POD('svd').fit(snapshots)
        snapshots_ = pod.expand(pod.reduce(snapshots))
        np.testing.assert_array_almost_equal(snapshots, snapshots_, decimal=4)

    def test_correlation_matrix_old(self):
        pod = POD('correlation_matrix').fit(snapshots)
        snapshots_ = pod.expand(pod.reduce(snapshots))
        np.testing.assert_array_almost_equal(snapshots, snapshots_, decimal=3)

    def test_correlation_matirix_savemem_old(self):
        pod = POD('correlation_matrix', save_memory=True).fit(snapshots)
        snapshots_ = pod.expand(pod.reduce(snapshots))
        np.testing.assert_array_almost_equal(snapshots, snapshots_, decimal=3)

    def test_randomized_svd_old(self):
        pod = POD('randomized_svd', save_memory=False).fit(snapshots)
        snapshots_ = pod.expand(pod.reduce(snapshots))
        np.testing.assert_array_almost_equal(snapshots, snapshots_, decimal=4)

    def test_singlular_values(self):
        a = POD('svd').fit(snapshots)
        np.testing.assert_allclose(
            a.singular_values,
            np.array([887.15704, 183.2508, 84.11757, 26.40448]),
            rtol=1e-6,
            atol=1e-8)

    def test_modes(self):
        a = POD('svd')
        a.fit(snapshots)
        np.testing.assert_allclose(a.modes, modes)

    def test_truncation_01(self):
        a = POD(method='svd', rank=0)
        a.fit(snapshots)
        assert a.singular_values.shape[0] == 1

    def test_truncation_02(self):
        a = POD(method='randomized_svd', rank=0)
        a.fit(snapshots)
        assert a.singular_values.shape[0] == 1

    def test_truncation_03(self):
        a = POD(method='correlation_matrix', rank=0)
        a.fit(snapshots)
        assert a.singular_values.shape[0] == 1

    def test_truncation_04(self):
        a = POD(method='svd', rank=3)
        a.fit(snapshots)
        assert a.singular_values.shape[0] == 3

    def test_truncation_05(self):
        a = POD(method='randomized_svd', rank=3)
        a.fit(snapshots)
        assert a.singular_values.shape[0] == 3

    def test_truncation_06(self):
        a = POD(method='correlation_matrix', rank=4)
        a.fit(snapshots)
        assert a.singular_values.shape[0] == 4

    def test_truncation_07(self):
        a = POD(method='svd', rank=0.8)
        a.fit(snapshots)
        assert a.singular_values.shape[0] == 1

    def test_truncation_08(self):
        a = POD(method='randomized_svd', rank=0.995)
        a.fit(snapshots)
        assert a.singular_values.shape[0] == 3

    def test_truncation_09(self):
        a = POD(method='correlation_matrix', rank=0.9999)
        a.fit(snapshots)
        assert a.singular_values.shape[0] == 4



import numpy as np
import warnings

from unittest import TestCase
from ezyrb import Linear, Database, POD, ReducedOrderModel

class TestKNeighbors(TestCase):
    def test_params(self):
        reg = Linear(fill_value=0)
        assert reg.fill_value == 0

    def test_default_params(self):
        reg = Linear()
        assert np.isnan(reg.fill_value)

    #def test_predict(self):
    #    reg = Linear()
    #    reg.fit([[1, 0, 0], [0, 1, 0], [0, 0, 1]], [[1, 0], [20, 5], [8, 6]])
    #    result = reg.predict([[1,2], [8,9], [6,7]])
    #    assert (result[0] == [1,0]).all()
    #    assert (result[1] == [8,6]).all()
    #    assert (result[2] == [20, 5]).all()

    def test_predict1d(self):
        reg = Linear()
        reg.fit([[1], [6], [8]], [[1, 0], [20, 5], [8, 6]])
        result = reg.predict([[1], [8], [6]])
        assert (result[0] == [1,0]).all()
        assert (result[1] == [8,6]).all()
        assert (result[2] == [20, 5]).all()

    def test_predict1d_2(self):
        reg = Linear()
        reg.fit([[1], [2]], [[1,1], [2,2]])
        result = reg.predict([[1.5]])
        assert (result[0] == [[1.5, 1.5]]).all()

    def test_predict1d_3(self):
        reg = Linear()
        reg.fit([1,2], [[1,1], [2,2]])
        result = reg.predict([[1.5]])
        assert (result[0] == [[1.5,1.5]]).all()

    def test_with_db_predict(self):
        reg = Linear()
        pod = POD()
        db = Database(np.array([1, 2, 3])[:,None], np.array([1, 5, 3])[:,None])
        rom = ReducedOrderModel(db, pod, reg)

        rom.fit()
        assert rom.predict([1]) == 1
        assert rom.predict([2]) == 5
        assert rom.predict([3]) == 3

    def test_wrong1(self):
        # wrong number of params
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", category=np.VisibleDeprecationWarning)
            with self.assertRaises(Exception):
                reg = Linear()
                reg.fit([[1, 2], [6,], [8, 9]], [[1, 0], [20, 5], [8, 6]])

    def test_wrong2(self):
        # wrong number of values
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", category=np.VisibleDeprecationWarning)
            with self.assertRaises(Exception):
                reg = Linear()
                reg.fit([[1, 2], [6,], [8, 9]], [[20, 5], [8, 6]])



"""EZyRB package"""

__all__ = [
    'Reduction', 'POD', 'Approximation', 'RBF', 'Linear', 'GPR',
    'ANN', 'KNeighborsRegressor', 'RadiusNeighborsRegressor', 'AE', 'AE_EDDL',
    'ReducedOrderModel'
]

from .reduction import Reduction
from .pod import POD
from .ae import AE
from .ae_eddl import AE_EDDL
from .approximation import Approximation
from .rbf import RBF
from .linear import Linear
from .gpr import GPR
from .reducedordermodel import ReducedOrderModel
from .ann import ANN
from .kneighbors_regressor import KNeighborsRegressor
from .radius_neighbors_regressor import RadiusNeighborsRegressor



"""
Module for FNN-Autoencoders.
"""

import torch
from .reduction import Reduction
from .ann import ANN


class AE(Reduction, ANN):
    """
    Feed-Forward AutoEncoder class (AE)

    :param list layers_encoder: ordered list with the number of neurons of
        each hidden layer for the encoder
    :param list layers_decoder: ordered list with the number of neurons of
        each hidden layer for the decoder
    :param torch.nn.modules.activation function_encoder: activation function
        at each layer for the encoder, except for the output layer at with
        Identity is considered by default.  A single activation function can
        be passed or a list of them of length equal to the number of hidden
        layers.
    :param torch.nn.modules.activation function_decoder: activation function
        at each layer for the decoder, except for the output layer at with
        Identity is considered by default.  A single activation function can
        be passed or a list of them of length equal to the number of hidden
        layers.
    :param list stop_training: list with the maximum number of training
        iterations (int) and/or the desired tolerance on the training loss
        (float).
    :param torch.nn.Module loss: loss definition (Mean Squared if not
        given).
    :param torch.optim optimizer: the torch class implementing optimizer.
        Default value is `Adam` optimizer.
    :param float lr: the learning rate. Default is 0.001.
    :param float l2_regularization: the L2 regularization coefficient, it
        corresponds to the "weight_decay". Default is 0 (no regularization).
    :param int frequency_print: the frequency in terms of epochs of the print
        during the training of the network.
    :param boolean last_identity: Flag to specify if the last activation
        function is the identity function. In the case the user provides the
        entire list of activation functions, this attribute is ignored. Default
        value is True.

    :Example:
        >>> from ezyrb import AE
        >>> import torch
        >>> f = torch.nn.Softplus
        >>> low_dim = 5
        >>> optim = torch.optim.Adam
        >>> ae = AE([400, low_dim], [low_dim, 400], f(), f(), 2000)
        >>> # or ...
        >>> ae = AE([400, 10, 10, low_dim], [low_dim, 400], f(), f(), 1e-5,
        >>>          optimizer=optim)
        >>> ae.fit(snapshots)
        >>> reduced_snapshots = ae.reduce(snapshots)
        >>> expanded_snapshots = ae.expand(reduced_snapshots)
    """
    def __init__(self,
                 layers_encoder,
                 layers_decoder,
                 function_encoder,
                 function_decoder,
                 stop_training,
                 loss=None,
                 optimizer=torch.optim.Adam,
                 lr=0.001,
                 l2_regularization=0,
                 frequency_print=10,
                 last_identity=True):

        if layers_encoder[-1] != layers_decoder[0]:
            raise ValueError('Wrong dimension in encoder and decoder layers')

        if loss is None:
            loss = torch.nn.MSELoss()

        if not isinstance(function_encoder, list):
            # Single activation function passed
            layers = layers_encoder
            nl = len(layers)-1 if last_identity else len(layers)
            function_encoder = [function_encoder] * nl

        if not isinstance(function_decoder, list):
            # Single activation function passed
            layers = layers_decoder
            nl = len(layers)-1 if last_identity else len(layers)
            function_decoder = [function_decoder] * nl

        if not isinstance(stop_training, list):
            stop_training = [stop_training]

        self.layers_encoder = layers_encoder
        self.layers_decoder = layers_decoder
        self.function_encoder = function_encoder
        self.function_decoder = function_decoder
        self.loss = loss

        self.stop_training = stop_training
        self.loss_trend = []
        self.encoder = None
        self.decoder = None
        self.optimizer = optimizer
        self.lr = lr
        self.frequency_print = frequency_print
        self.l2_regularization = l2_regularization

    def _build_model(self, values):
        """
        Build the torch model.

        Considering the number of neurons per layer (self.layers), a
        feed-forward NN is defined:
            - activation function from layer i>=0 to layer i+1:
              self.function[i]; activation function at the output layer:
              Identity (by default).

        :param numpy.ndarray values: the set values one wants to reduce.
        """
        layers_encoder = self.layers_encoder.copy()
        layers_encoder.insert(0, values.shape[1])
        self.encoder = self._list_to_sequential(layers_encoder,
                                                self.function_encoder)

        layers_decoder = self.layers_decoder.copy()
        layers_decoder.append(values.shape[1])
        self.decoder = self._list_to_sequential(layers_decoder,
                                                self.function_decoder)

    def fit(self, values):
        """
        Build the AE given 'values' and perform training.

        Training procedure information:
            -  optimizer: Adam's method with default parameters (see, e.g.,
               https://pytorch.org/docs/stable/optim.html);
            -  loss: self.loss (if none, the Mean Squared Loss is set by
               default).
            -  stopping criterion: the fulfillment of the requested tolerance
               on the training loss compatibly with the prescribed budget of
               training iterations (if type(self.stop_training) is list); if
               type(self.stop_training) is int or type(self.stop_training) is
               float, only the number of maximum iterations or the accuracy
               level on the training loss is considered as the stopping rule,
               respectively.

        :param numpy.ndarray values: the (training) values in the points.
        """
        values = values.T
        self._build_model(values)

        optimizer = self.optimizer(
            list(self.encoder.parameters()) + list(self.decoder.parameters()),
            lr=self.lr, weight_decay=self.l2_regularization)

        values = self._convert_numpy_to_torch(values)

        n_epoch = 1
        flag = True
        while flag:
            y_pred = self.decoder(self.encoder(values))

            loss = self.loss(y_pred, values)

            optimizer.zero_grad()
            loss.backward()
            optimizer.step()

            scalar_loss = loss.item()
            self.loss_trend.append(scalar_loss)

            for criteria in self.stop_training:
                if isinstance(criteria, int):  # stop criteria is an integer
                    if n_epoch == criteria:
                        flag = False
                elif isinstance(criteria, float):  # stop criteria is float
                    if scalar_loss < criteria:
                        flag = False

            if (flag is False or
                    n_epoch == 1 or n_epoch % self.frequency_print == 0):
                print(f'[epoch {n_epoch:6d}]\t{scalar_loss:e}')

            n_epoch += 1

        return optimizer

    def transform(self, X):
        """
        Reduces the given snapshots.

        :param numpy.ndarray X: the input snapshots matrix (stored by column).
        """
        X = self._convert_numpy_to_torch(X).T
        g = self.encoder(X)
        return g.cpu().detach().numpy().T

    def inverse_transform(self, g):
        """
        Projects a reduced to full order solution.

        :param: numpy.ndarray g the latent variables.
        """
        g = self._convert_numpy_to_torch(g).T
        u = self.decoder(g)
        return u.cpu().detach().numpy().T

    def reduce(self, X):
        """
        Reduces the given snapshots.

        :param numpy.ndarray X: the input snapshots matrix (stored by column).

        .. note::

            Same as `transform`. Kept for backward compatibility.
        """
        return self.transform(X)

    def expand(self, g):
        """
        Projects a reduced to full order solution.

        :param: numpy.ndarray g the latent variables.

        .. note::

            Same as `inverse_transform`. Kept for backward compatibility.
        """
        return self.inverse_transform(g)



import numpy as np
from unittest import TestCase, main
from ezyrb import RegularGrid, Database, POD, ReducedOrderModel


class TestRegularGrid(TestCase):

    def test_1D_1mode(self):
        reg = RegularGrid()
        reg.fit(np.array([1, 2, 3])[:, None],
                np.array([1, 5, 3])[:, None])
        assert reg.predict([1]) == 1
        assert reg.predict([2]) == 5
        assert reg.predict([3]) == 3

    def test_predict1d(self):
        reg = RegularGrid()
        x1 = np.array([[1], [6], [8]])
        V = np.array([[1, 0], [20, 5], [8, 6]])  # n, r = 3, 2
        reg.fit(x1, V)
        result = reg.predict([[1], [8], [6]])
        assert (result[0] == [1, 0]).all()
        assert (result[1] == [8, 6]).all()
        assert (result[2] == [20, 5]).all()

    def test_predict1d_2(self):
        reg = RegularGrid()
        x1 = [1, 2]
        reg.fit(x1, [[1, 1], [2, 2]])
        result = reg.predict([[1.5]])
        assert (result[0] == [[1.5, 1.5]]).all()

    def test_predict2d(self):
        reg = RegularGrid()
        x1 = [1, 2, 3]
        x2 = [4, 5, 6, 7]
        xx1, xx2 = np.meshgrid(x1, x2, indexing="ij")
        points = np.c_[xx1.ravel(), xx2.ravel()]
        V = [[1, 21], [2, 22], [3, 23], [4, 24], [5, 25], [6, 26],
             [7, 27], [8, 28], [9, 29], [10, 30], [11, 31], [12, 32]]
        reg.fit(points, V, method="linear")
        result = reg.predict([[1, 4], [1, 5]])
        assert (result[0] == [1, 21]).all()
        assert (result[1] == [2, 22]).all()

    def test_get_grid_axes_2D(self):
        x1 = [.1, .2, .3]
        x2 = [4, 5, 6, 7]
        xx1, xx2 = np.meshgrid(x1, x2, indexing="ij")
        V = np.arange(len(x1)*len(x2)*2).reshape(-1, 2)
        pts = np.c_[xx1.ravel(), xx2.ravel()]

        random_order = np.arange(len(pts))
        np.random.shuffle(random_order)
        pts_scrmbld = pts[random_order, :]
        V_scrmbld = V[random_order, :]

        reg = RegularGrid()
        grid_axes, V_unscrambled = reg.get_grid_axes(pts_scrmbld, V_scrmbld)

        assert np.allclose(V_unscrambled, V)
        assert np.allclose(grid_axes[0], x1)
        assert np.allclose(grid_axes[1], x2)

    def test_get_grid_axes_3D(self):
        x1 = [.1, .2, .3]
        x2 = [4, 5, 6, 7]
        x3 = [.12, .34, .56, .78, .90]
        xx1, xx2, xx3 = np.meshgrid(x1, x2, x3, indexing="ij")
        V = np.arange(len(x1)*len(x2)*len(x3)*2).reshape(-1, 2)
        pts = np.c_[xx1.ravel(), xx2.ravel(), xx3.ravel()]

        random_order = np.arange(len(pts))
        np.random.shuffle(random_order)
        pts_scrmbld = pts[random_order, :]
        V_scrmbld = V[random_order, :]

        reg = RegularGrid()
        grid_axes, V_unscrambled = reg.get_grid_axes(pts_scrmbld, V_scrmbld)

        assert np.allclose(V_unscrambled, V)
        assert np.allclose(grid_axes[0], x1)
        assert np.allclose(grid_axes[1], x2)
        assert np.allclose(grid_axes[2], x3)

    def test_with_db_predict(self):
        reg = RegularGrid()
        pod = POD()
        db = Database(np.array([1, 2, 3])[:, None],
                      np.array([1, 5, 3])[:, None])
        rom = ReducedOrderModel(db, pod, reg)
        rom.fit()
        assert rom.predict([1]) == 1
        assert rom.predict([2]) == 5
        assert rom.predict([3]) == 3

    def test_fails(self):
        reg = RegularGrid()
        reg = RegularGrid()
        p = [[1, 2]]
        V = [[1, 1], [2, 2]]
        with self.assertRaises(ValueError):
            reg.fit(p, V)


if __name__ == "__main__":
    main()



"""
Module for higher order interpolation on regular grids
"""
import numpy as np
from scipy.interpolate import RegularGridInterpolator as RGI
from .approximation import Approximation


class RegularGrid(Approximation):
    """
    Multidimensional interpolator on regular grids.
    See also scipy's RegularGridInterpolator for information on kwargs.

    :param array_like points: the coordinates of the points on a regular grid.
    :param array_like values: The (vector-) data on the regular grid in
        n dimensions.

    :Example:

        >>> import ezyrb
        >>> import numpy as np
        >>> def f(x, y, z):
        ...     return 2 * x**3 + 3 * y**2 - z
        >>> x = np.linspace(1, 4, 11)
        >>> y = np.linspace(4, 7, 22)
        >>> z = np.linspace(7, 9, 33)
        >>> xg, yg, zg = np.meshgrid(x, y, z, indexing='ij')
        >>> points = np.c_[xg.ravel(), yg.ravel(), zg.ravel()]
        >>> data_mode_x = f(xg, yg, zg).reshape(-1, 1)
        # lets assume we have 2 modes, i.e. a rank 2 model
        >>> data = np.concatenate((data_mode_x, data_mode_x/10), axis=1)
        >>> rgi = ezyrb.RegularGrid()
        >>> rgi.fit(points, data, method="linear")
        >>> pts = np.array([[2.1, 6.2, 8.3],
        ...                 [3.3, 5.2, 7.1],
        ...                 [1., 4., 7.],
        ...                 [4., 7., 9.]])
        >>> rgi.predict(pts)
        array([[125.80469388,  12.58046939],
               [146.30069388,  14.63006939],
               [ 43.        ,   4.3       ],
               [266.        ,  26.6       ]])
        >>> f(pts[:, 0], pts[:, 1], pts[:, 2])
        array([125.542, 145.894,  43.   , 266.   ])
        >>> f(pts[:, 0], pts[:, 1], pts[:, 2])/10
        array([12.5542, 14.5894,  4.3   , 26.6   ])
    """

    def __init__(self):
        self.interpolator = None

    def get_grid_axes(self, pts_scrmbld, vals_scrmbld):
        """
        Calculates the grid axes from a meshed grid and re-orders the values so
        they fit on a mesh generated with
        numpy.meshgrid(ax1, ax2, ..., axn, indexing="ij")

        :param array_like pts_scrmbld: the coordinates of the points.
        :param array_like vals_scrmbld: the (vector-)values in the points.
        :return: The grid axes given as a tuple of ndarray, with shapes
            (m1, ), …, (mn, ) and values mapped on the ordered grid.
        :rtype: (list, numpy.ndarray)

        """
        grid_axes = []
        iN = []  # index in dimension N
        nN = []  # size of dimension N
        dim = pts_scrmbld.shape[1]
        for i in range(dim):
            xn, unique_inverse_n = np.unique(pts_scrmbld[:, i],
                                             return_inverse=True)
            grid_axes.append(xn)
            nN.append(len(xn))
            iN.append(unique_inverse_n)
        if np.prod(nN) != len(vals_scrmbld):
            msg = "Values don't match grid. Make sure to pass a list of "\
                  "points that are on a regular grid! Be aware of floating "\
                  "point precision."
            raise ValueError(msg)
        new_row_index = np.ravel_multi_index(iN, nN)
        reverse_scrambling = np.argsort(new_row_index)
        vals_on_regular_grid = vals_scrmbld[reverse_scrambling]
        return grid_axes, vals_on_regular_grid

    def fit(self, points, values, **kwargs):
        """
        Construct the interpolator given `points` and `values`.
        Assumes that the points are on a regular grid, fails when not.
        see scipy.interpolate.RegularGridInterpolator

        :param array_like points: the coordinates of the points.
        :param array_like values: the values in the points.
        """
        points = np.asarray(points)
        if not np.issubdtype(points.dtype, np.number):
            raise ValueError('Invalid format or dimension for the argument'
                             '`points`.')
        if points.ndim == 1:
            points = points[:, None]
        vals = np.asarray(values)
        grid_axes, values_grd = self.get_grid_axes(points, vals)
        shape = [len(ax) for ax in grid_axes]
        shape.append(-1)
        self.interpolator = RGI(grid_axes, values_grd.reshape(shape), **kwargs)

    def predict(self, new_point):
        """
        Evaluate interpolator at given `new_point`, can be multiple points.

        :param array_like new_point: the coordinates of the given point(s).
        :return: the interpolated values.
        :rtype: numpy.ndarray
        """
        new_point = np.asarray(new_point)
        if new_point.ndim == 1:
            new_point = new_point[:, None]
        return self.interpolator(new_point)



"""
Module for Artificial Neural Network (ANN) Prediction.
"""

import torch
import torch.nn as nn
import numpy as np
from .approximation import Approximation


class ANN(Approximation):
    """
    Feed-Forward Artifical Neural Network (ANN).

    :param list layers: ordered list with the number of neurons of each hidden
        layer.
    :param torch.nn.modules.activation function: activation function at each
        layer. A single activation function can be passed or a list of them of
        length equal to the number of hidden layers.
    :param list stop_training: list with the maximum number of training
        iterations (int) and/or the desired tolerance on the training loss
        (float).
    :param torch.nn.Module loss: loss definition (Mean Squared if not given).
    :param torch.optim optimizer: the torch class implementing optimizer.
        Default value is `Adam` optimizer.
    :param float lr: the learning rate. Default is 0.001.
    :param float l2_regularization: the L2 regularization coefficient, it
        corresponds to the "weight_decay". Default is 0 (no regularization).
    :param int frequency_print: the frequency in terms of epochs of the print
        during the training of the network.
    :param boolean last_identity: Flag to specify if the last activation
        function is the identity function. In the case the user provides the
        entire list of activation functions, this attribute is ignored. Default
        value is True.

    :Example:
        >>> import ezyrb
        >>> import numpy as np
        >>> import torch.nn as nn
        >>> x = np.random.uniform(-1, 1, size =(4, 2))
        >>> y = np.array([np.sin(x[:, 0]), np.cos(x[:, 1]**3)]).T
        >>> ann = ezyrb.ANN([10, 5], nn.Tanh(), [20000,1e-5])
        >>> ann.fit(x, y)
        >>> y_pred = ann.predict(x)
        >>> print(y)
        >>> print(y_pred)
        >>> print(len(ann.loss_trend))
        >>> print(ann.loss_trend[-1])
    """
    def __init__(self, layers, function, stop_training, loss=None,
                 optimizer=torch.optim.Adam, lr=0.001, l2_regularization=0,
                 frequency_print=10, last_identity=True):
        if loss is None:
            loss = torch.nn.MSELoss()

        if not isinstance(function, list):  # Single activation function passed
            nl = len(layers) if last_identity else len(layers)+1
            function = [function] * nl

        if not isinstance(stop_training, list):
            stop_training = [stop_training]

        self.layers = layers
        self.function = function
        self.loss = loss
        self.stop_training = stop_training

        self.loss_trend = []
        self.model = None
        self.optimizer = optimizer

        self.frequency_print = frequency_print
        self.lr = lr
        self.l2_regularization = l2_regularization

    def _convert_numpy_to_torch(self, array):
        """
        Converting data type.

        :param numpy.ndarray array: input array.
        :return: the tensorial counter-part of the input array.
        :rtype: torch.Tensor.
        """
        return torch.from_numpy(array).float()

    def _convert_torch_to_numpy(self, tensor):
        """
        Converting data type.

        :param torch.Tensor tensor: input tensor.
        :return: the vectorial counter-part of the input tensor.
        :rtype: numpy.ndarray.
        """
        return tensor.detach().numpy()

    @staticmethod
    def _list_to_sequential(layers, functions):

        layers_torch = []
        inout_layers = [[layers[i], layers[i+1]] for i in range(len(layers)-1)]

        while True:
            if inout_layers:
                inp_d, out_d = inout_layers.pop(0)
                layers_torch.append(nn.Linear(inp_d, out_d))

            if functions:
                layers_torch.append(functions.pop(0))

            if not functions and not inout_layers:
                break

        return nn.Sequential(*layers_torch)

    def _build_model(self, points, values):
        """
        Build the torch model.
        Considering the number of neurons per layer (self.layers), a
        feed-forward NN is defined:
        -  activation function from layer i>=0 to layer i+1:
           self.function[i]; activation function at the output layer:
           Identity (by default).
        :param numpy.ndarray points: the coordinates of the given (training)
        points.
        :param numpy.ndarray values: the (training) values in the points.
        """
        layers = self.layers.copy()
        layers.insert(0, points.shape[1])
        layers.append(values.shape[1])

        self.model = self._list_to_sequential(layers, self.function)

    def fit(self, points, values):
        """
        Build the ANN given 'points' and 'values' and perform training.

        Training procedure information:
            -  optimizer: Adam's method with default parameters (see, e.g.,
               https://pytorch.org/docs/stable/optim.html);
            -  loss: self.loss (if none, the Mean Squared Loss is set by
               default).
            -  stopping criterion: the fulfillment of the requested tolerance
               on the training loss compatibly with the prescribed budget of
               training iterations (if type(self.stop_training) is list); if
               type(self.stop_training) is int or type(self.stop_training) is
               float, only the number of maximum iterations or the accuracy
               level on the training loss is considered as the stopping rule,
               respectively.

        :param numpy.ndarray points: the coordinates of the given (training)
            points.
        :param numpy.ndarray values: the (training) values in the points.
        """

        self._build_model(points, values)
        optimizer = self.optimizer(
            self.model.parameters(),
            lr=self.lr, weight_decay=self.l2_regularization)

        points = self._convert_numpy_to_torch(points)
        values = self._convert_numpy_to_torch(values)

        n_epoch = 1
        flag = True
        while flag:
            y_pred = self.model(points)

            loss = self.loss(y_pred, values)

            optimizer.zero_grad()
            loss.backward()
            optimizer.step()

            scalar_loss = loss.item()
            self.loss_trend.append(scalar_loss)

            for criteria in self.stop_training:
                if isinstance(criteria, int):  # stop criteria is an integer
                    if n_epoch == criteria:
                        flag = False
                elif isinstance(criteria, float):  # stop criteria is float
                    if scalar_loss < criteria:
                        flag = False

            if (flag is False or
                    n_epoch == 1 or n_epoch % self.frequency_print == 0):
                print(f'[epoch {n_epoch:6d}]\t{scalar_loss:e}')

            n_epoch += 1

        return optimizer

    def predict(self, new_point):
        """
        Evaluate the ANN at given 'new_points'.

        :param array_like new_points: the coordinates of the given points.
        :return: the predicted values via the ANN.
        :rtype: numpy.ndarray
        """
        new_point = self._convert_numpy_to_torch(np.array(new_point))
        y_new = self.model(new_point)
        return self._convert_torch_to_numpy(y_new)
